'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var path = require('path');
var core = require('@babel/core');
var template = require('@babel/template');
var findUp = require('find-up');

function _interopDefault (e) { return e && e.__esModule ? e : { 'default': e }; }

var template__default = /*#__PURE__*/_interopDefault(template);
var findUp__default = /*#__PURE__*/_interopDefault(findUp);

const packageIdentifier = '@vanilla-extract/css';
const buildSetFileScope = template__default['default'](`
  import { setFileScope, endFileScope } from %%packageIdentifier%%
  setFileScope(%%filePath%%, %%packageName%%)
`);
const debuggableFunctionConfig = {
  style: {
    maxParams: 2
  },
  createTheme: {
    maxParams: 3
  },
  mapToStyles: {
    maxParams: 3
  },
  fontFace: {
    maxParams: 2
  },
  keyframes: {
    maxParams: 2
  },
  createVar: {
    maxParams: 1
  }
};
const styleFunctions = [...Object.keys(debuggableFunctionConfig), 'globalStyle', 'createGlobalTheme', 'createThemeVars', 'globalFontFace', 'globalKeyframes'];

const extractName = node => {
  if (core.types.isObjectProperty(node) && core.types.isIdentifier(node.key)) {
    return node.key.name;
  } else if ((core.types.isVariableDeclarator(node) || core.types.isFunctionDeclaration(node)) && core.types.isIdentifier(node.id)) {
    return node.id.name;
  } else if (core.types.isExportDefaultDeclaration(node)) {
    return 'default';
  } else if (core.types.isVariableDeclarator(node) && core.types.isArrayPattern(node.id) && core.types.isIdentifier(node.id.elements[0])) {
    return node.id.elements[0].name;
  }
};

const getDebugId = path => {
  const {
    parent
  } = path;

  if (core.types.isObjectProperty(parent) || core.types.isReturnStatement(parent) || core.types.isArrowFunctionExpression(parent) || core.types.isArrayExpression(parent) || core.types.isSpreadElement(parent)) {
    const names = [];
    path.findParent(({
      node: parentNode
    }) => {
      const name = extractName(parentNode);

      if (name) {
        names.unshift(name);
      } // Traverse all the way to the root


      return false;
    });
    return names.join('_');
  } else {
    return extractName(parent);
  }
};

const getRelevantCall = (node, namespaceImport, importIdentifiers) => {
  const {
    callee
  } = node;

  if (namespaceImport && core.types.isMemberExpression(callee) && core.types.isIdentifier(callee.object, {
    name: namespaceImport
  })) {
    return styleFunctions.find(exportName => core.types.isIdentifier(callee.property, {
      name: exportName
    }));
  } else {
    const importInfo = Array.from(importIdentifiers.entries()).find(([identifier]) => core.types.isIdentifier(callee, {
      name: identifier
    }));

    if (importInfo) {
      return importInfo[1];
    }
  }
};

function index () {
  let packageInfo;
  let hasResolvedPackageJson = false;

  function getPackageInfo(cwd) {
    if (hasResolvedPackageJson) {
      return packageInfo;
    }

    hasResolvedPackageJson = true;
    const packageJsonPath = findUp__default['default'].sync('package.json', {
      cwd: cwd || undefined
    });

    if (packageJsonPath) {
      const {
        name
      } = require(packageJsonPath);

      packageInfo = {
        name,
        path: packageJsonPath,
        dirname: path.dirname(packageJsonPath)
      };
    }

    return packageInfo;
  }

  return {
    pre({
      opts
    }) {
      this.importIdentifiers = new Map();
      this.namespaceImport = '';

      if (!opts.filename) {
        // TODO Make error better
        throw new Error('Filename must be available');
      }

      const packageInfo = getPackageInfo(opts.cwd);

      if (!packageInfo) {
        throw new Error(`Couldn't find parent package.json for ${opts.filename}`);
      }

      if (!packageInfo.name) {
        throw new Error(`Closest package.json (${packageInfo.path}) must specify name`);
      }

      this.packageName = packageInfo.name;
      this.filePath = path.relative(packageInfo.dirname, opts.filename);
    },

    visitor: {
      Program: {
        exit(path) {
          if (this.importIdentifiers.size > 0 || this.namespaceImport) {
            // Wrap module with file scope calls
            path.unshiftContainer('body', buildSetFileScope({
              packageIdentifier: core.types.stringLiteral(`${packageIdentifier}/fileScope`),
              filePath: core.types.stringLiteral(this.filePath),
              packageName: core.types.stringLiteral(this.packageName)
            }));
            path.pushContainer('body', core.types.callExpression(core.types.identifier('endFileScope'), []));
          }
        }

      },

      ImportDeclaration(path) {
        if (path.node.source.value === packageIdentifier) {
          path.node.specifiers.forEach(specifier => {
            if (core.types.isImportNamespaceSpecifier(specifier)) {
              this.namespaceImport = specifier.local.name;
            } else if (core.types.isImportSpecifier(specifier)) {
              const {
                imported,
                local
              } = specifier;
              const importName = core.types.isIdentifier(imported) ? imported.name : imported.value;

              if (styleFunctions.includes(importName)) {
                this.importIdentifiers.set(local.name, importName);
              }
            }
          });
        }
      },

      CallExpression(path) {
        const {
          node
        } = path;
        const usedExport = getRelevantCall(node, this.namespaceImport, this.importIdentifiers);

        if (usedExport && usedExport in debuggableFunctionConfig) {
          if (node.arguments.length < debuggableFunctionConfig[usedExport].maxParams) {
            const debugIdent = getDebugId(path);

            if (debugIdent) {
              node.arguments.push(core.types.stringLiteral(debugIdent));
            }
          }
        }
      }

    }
  };
}

exports.default = index;
