import { relative, dirname } from 'path';
import { types } from '@babel/core';
import template from '@babel/template';
import findUp from 'find-up';

const packageIdentifier = '@vanilla-extract/css';
const buildSetFileScope = template(`
  import { setFileScope, endFileScope } from %%packageIdentifier%%
  setFileScope(%%filePath%%, %%packageName%%)
`);
const debuggableFunctionConfig = {
  style: {
    maxParams: 2
  },
  createTheme: {
    maxParams: 3
  },
  mapToStyles: {
    maxParams: 3
  },
  fontFace: {
    maxParams: 2
  },
  keyframes: {
    maxParams: 2
  },
  createVar: {
    maxParams: 1
  }
};
const styleFunctions = [...Object.keys(debuggableFunctionConfig), 'globalStyle', 'createGlobalTheme', 'createThemeVars', 'globalFontFace', 'globalKeyframes'];

const extractName = node => {
  if (types.isObjectProperty(node) && types.isIdentifier(node.key)) {
    return node.key.name;
  } else if ((types.isVariableDeclarator(node) || types.isFunctionDeclaration(node)) && types.isIdentifier(node.id)) {
    return node.id.name;
  } else if (types.isExportDefaultDeclaration(node)) {
    return 'default';
  } else if (types.isVariableDeclarator(node) && types.isArrayPattern(node.id) && types.isIdentifier(node.id.elements[0])) {
    return node.id.elements[0].name;
  }
};

const getDebugId = path => {
  const {
    parent
  } = path;

  if (types.isObjectProperty(parent) || types.isReturnStatement(parent) || types.isArrowFunctionExpression(parent) || types.isArrayExpression(parent) || types.isSpreadElement(parent)) {
    const names = [];
    path.findParent(({
      node: parentNode
    }) => {
      const name = extractName(parentNode);

      if (name) {
        names.unshift(name);
      } // Traverse all the way to the root


      return false;
    });
    return names.join('_');
  } else {
    return extractName(parent);
  }
};

const getRelevantCall = (node, namespaceImport, importIdentifiers) => {
  const {
    callee
  } = node;

  if (namespaceImport && types.isMemberExpression(callee) && types.isIdentifier(callee.object, {
    name: namespaceImport
  })) {
    return styleFunctions.find(exportName => types.isIdentifier(callee.property, {
      name: exportName
    }));
  } else {
    const importInfo = Array.from(importIdentifiers.entries()).find(([identifier]) => types.isIdentifier(callee, {
      name: identifier
    }));

    if (importInfo) {
      return importInfo[1];
    }
  }
};

function index () {
  let packageInfo;
  let hasResolvedPackageJson = false;

  function getPackageInfo(cwd) {
    if (hasResolvedPackageJson) {
      return packageInfo;
    }

    hasResolvedPackageJson = true;
    const packageJsonPath = findUp.sync('package.json', {
      cwd: cwd || undefined
    });

    if (packageJsonPath) {
      const {
        name
      } = require(packageJsonPath);

      packageInfo = {
        name,
        path: packageJsonPath,
        dirname: dirname(packageJsonPath)
      };
    }

    return packageInfo;
  }

  return {
    pre({
      opts
    }) {
      this.importIdentifiers = new Map();
      this.namespaceImport = '';

      if (!opts.filename) {
        // TODO Make error better
        throw new Error('Filename must be available');
      }

      const packageInfo = getPackageInfo(opts.cwd);

      if (!packageInfo) {
        throw new Error(`Couldn't find parent package.json for ${opts.filename}`);
      }

      if (!packageInfo.name) {
        throw new Error(`Closest package.json (${packageInfo.path}) must specify name`);
      }

      this.packageName = packageInfo.name;
      this.filePath = relative(packageInfo.dirname, opts.filename);
    },

    visitor: {
      Program: {
        exit(path) {
          if (this.importIdentifiers.size > 0 || this.namespaceImport) {
            // Wrap module with file scope calls
            path.unshiftContainer('body', buildSetFileScope({
              packageIdentifier: types.stringLiteral(`${packageIdentifier}/fileScope`),
              filePath: types.stringLiteral(this.filePath),
              packageName: types.stringLiteral(this.packageName)
            }));
            path.pushContainer('body', types.callExpression(types.identifier('endFileScope'), []));
          }
        }

      },

      ImportDeclaration(path) {
        if (path.node.source.value === packageIdentifier) {
          path.node.specifiers.forEach(specifier => {
            if (types.isImportNamespaceSpecifier(specifier)) {
              this.namespaceImport = specifier.local.name;
            } else if (types.isImportSpecifier(specifier)) {
              const {
                imported,
                local
              } = specifier;
              const importName = types.isIdentifier(imported) ? imported.name : imported.value;

              if (styleFunctions.includes(importName)) {
                this.importIdentifiers.set(local.name, importName);
              }
            }
          });
        }
      },

      CallExpression(path) {
        const {
          node
        } = path;
        const usedExport = getRelevantCall(node, this.namespaceImport, this.importIdentifiers);

        if (usedExport && usedExport in debuggableFunctionConfig) {
          if (node.arguments.length < debuggableFunctionConfig[usedExport].maxParams) {
            const debugIdent = getDebugId(path);

            if (debugIdent) {
              node.arguments.push(types.stringLiteral(debugIdent));
            }
          }
        }
      }

    }
  };
}

export default index;
